/* Code to get discid for a cddb query. */

/*

  Based on code found in:

  To: code-review@azure.humbug.org.au 
  Subject: CDDB database reader 
  From: Byron Ellacott <rodent@route-qn.uqnga.org.au> 
  Date: Fri, 5 Jun 1998 17:32:40 +1000 

*/

/* Stripped net code, 'cause I only care about the discid */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/cdio.h>

#include "cdflags.h"
#include "discid.h"
#include "config.h"

const char *cdrom_device=DEV_CDROM;
int cdrom_fd;

int Discid_Init(void) {
/*	if ((cdrom_device = getenv("CD_DEVICE")) == NULL) {
		cdrom_device = "/dev/cdrom";
	}
*/
    cdrom_fd = open(cdrom_device, O_RDONLY | O_NONBLOCK);
	return(0);
}

struct toc {
    int min, sec, frame;
} cdtoc[100];

int read_toc(void) {
    struct ioc_toc_header tochdr;
    struct cd_toc_entry *tocentry;
    struct ioc_read_toc_entry ioc_toc;
    int i,len,first,last;

    ioctl(cdrom_fd, CDIOREADTOCHEADER, (char*)&tochdr);

    first=tochdr.starting_track;
    last=tochdr.ending_track;
    len=(last-first+2)*sizeof(struct cd_toc_entry);
    tocentry=(struct cd_toc_entry*)malloc(len);

    ioc_toc.starting_track=0;
    ioc_toc.address_format=CD_MSF_FORMAT;
    ioc_toc.data_len=len;
    ioc_toc.data=tocentry;

    ioctl(cdrom_fd,CDIOREADTOCENTRYS,(char*)&ioc_toc);

    for (i = first; i <= last; i++) {
        cdtoc[i-1].min = ioc_toc.data[i-1].addr.msf.minute;
        cdtoc[i-1].sec = ioc_toc.data[i-1].addr.msf.second;
        cdtoc[i-1].frame = ioc_toc.data[i-1].addr.msf.frame;
        cdtoc[i-1].frame += cdtoc[i-1].min*60*75;
        cdtoc[i-1].frame += cdtoc[i-1].sec*75;
    }
    cdtoc[last].min = ioc_toc.data[last].addr.msf.minute;
    cdtoc[last].sec = ioc_toc.data[last].addr.msf.second;
    cdtoc[last].frame = ioc_toc.data[last].addr.msf.frame;
    cdtoc[last].frame += cdtoc[last].min*60*75;
    cdtoc[last].frame += cdtoc[last].sec*75;

    ioc_toc.starting_track = DISCID_TRACK;
    ioc_toc.data_len=sizeof(struct cd_toc_entry);
    ioctl(cdrom_fd, CDIOREADTOCENTRYS, &tocentry);

    close(cdrom_fd);

    return last;
}

unsigned int cddb_sum(int n) {
    unsigned int ret;

    ret = 0;
    while (n > 0) {
      ret += (n % 10);
      n /= 10;
    }
    return ret;
}

unsigned long cddb_discid(unsigned long tot_trks) {
    unsigned int i, t = 0, n = 0;

    i = 0;
    while (i < tot_trks) {
      n = n + cddb_sum((cdtoc[i].min * 60) + cdtoc[i].sec);
      i++;
    }
    t = ((cdtoc[tot_trks].min * 60) + cdtoc[tot_trks].sec) -
      ((cdtoc[0].min * 60) + cdtoc[0].sec);
    return ((n % 0xff) << 24 | t << 8 | tot_trks);
}

int discid(void) {
    unsigned long discid;
    int tracks, i;

    tracks = read_toc();
    discid = cddb_discid(tracks);
    printf("%08x %d", discid, tracks);
    for (i = 0; i < tracks; i++) printf(" %d", cdtoc[i].frame);
    printf(" %d\n", (cdtoc[tracks].frame-cdtoc[0].frame)/75);
	//printf("\n");
    return(0);
}



