#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/time.h>
#include <sys/param.h>
#include <signal.h>
#include <ustat.h>
#include <poll.h>
#include <sys/cdio.h>
#include <sys/socket.h>
#include <sys/scsi/impl/uscsi.h>

#include "sun.h"
#include "taksignal.h"
#include "cdflags.h"
#include "config.h"

int list_tracks(int cdrom_fd){
	int returnvalue=0,tracklength=0;
	struct cdrom_tocentry toc;

	toc.cdte_format=CDROM_MSF;
	toc.cdte_track=2;

	while(ioctl(cdrom_fd,CDROMREADTOCENTRY,&toc)>=0){
		tracklength=60*toc.cdte_addr.msf.minute+toc.cdte_addr.msf.second+1-tracklength;
		printf("Track %d: %d seconds\n",toc.cdte_track-1,tracklength);
		toc.cdte_track++;
	}//while
	
	toc.cdte_track=CDROM_LEADOUT;

	if(ioctl(cdrom_fd,CDROMREADTOCENTRY,&toc)>=0){
		tracklength=60*toc.cdte_addr.msf.minute+toc.cdte_addr.msf.second+1-tracklength;
		printf("Track %d: %d seconds\n",toc.cdte_track-1,tracklength);
	}//if
	else{
		 fprintf(stderr,"Error reading tracks on device %s\n",DEV_CDROM);
		returnvalue=-1;
	}//else

	return returnvalue;
}//list_tracks


int play_track(int cdrom_fd,int track,int *seconds,time_t *starttime){
	int returnvalue=0;
	struct cdrom_tocentry toc;
	struct cdrom_msf msf;

	printf("Playing track %d\n",track);

	toc.cdte_format=CDROM_MSF;
	toc.cdte_track=track;

	if((ioctl(cdrom_fd,CDROMREADTOCENTRY,&toc))<0){
		fprintf(stderr,"Error reading table of contents\n");
		returnvalue=-1;
	}//if

	msf.cdmsf_min0=toc.cdte_addr.msf.minute;
	msf.cdmsf_sec0=toc.cdte_addr.msf.second;
	msf.cdmsf_frame0=toc.cdte_addr.msf.frame;
	toc.cdte_track++;

	if((ioctl(cdrom_fd,CDROMREADTOCENTRY,&toc))<0){
		printf("Trying leadout...\n");
		toc.cdte_track=CDROM_LEADOUT;
		if(ioctl(cdrom_fd,CDROMREADTOCENTRY,&toc)<0){
			fprintf(stderr,"Error reading table of contents\n");
			returnvalue=-1;
		}//if
	}//if
	msf.cdmsf_min1=toc.cdte_addr.msf.minute;
	msf.cdmsf_sec1=toc.cdte_addr.msf.second;
	msf.cdmsf_frame1=toc.cdte_addr.msf.frame;

	(*seconds)=((msf.cdmsf_min1-msf.cdmsf_min0)*60+(msf.cdmsf_sec1-msf.cdmsf_sec0)+1);
	(*starttime)=time(NULL);
		
	if(ioctl(cdrom_fd,CDROMPLAYMSF,&msf)==-1){
		fprintf(stderr,"Error playing track %d\n",track);
			returnvalue=-1;
	}//if

	printf("Track %d: %d seconds\n",track,*seconds);
		
	sleep((*seconds));

	return returnvalue;
}//play_track



int stop_cd(int cdrom_fd){
	int returnvalue=0;

	if(ioctl(cdrom_fd,CDROMSTOP,NULL)==-1){
		fprintf(stderr,"Error stopping CD\n");
		returnvalue=-1;
	}//if

	return returnvalue;
}//stop_cd
		

int resume_cd(int cdrom_fd,int *seconds,time_t *starttime,time_t stoptime){
	int returnvalue=0;
	time_t temp=time(NULL);
//	NEVER drink & code...or is that ALWAYS?
//	printf("Started at %d, paused at %d - %d left to play\n",(*starttime),stoptime,(*seconds)-(stoptime-(*starttime)));
	(*seconds)-=(int)(stoptime-(*starttime));
	(*starttime)=temp;

	if(ioctl(cdrom_fd,CDROMRESUME,NULL)==-1){
//		fprintf(stderr,"Error resuming cd play\n");
		returnvalue=-1;
	}//if
	else sleep(*seconds);
	
	return returnvalue;
}//resume_cd


int pause_cd(int cdrom_fd,time_t *stoptime){
	int returnvalue=0;

	(*stoptime)=time(NULL);

	if(ioctl(cdrom_fd,CDROMPAUSE,NULL)==-1)
		returnvalue=-1;

	pause();

	return returnvalue;
}//pause_cd

