#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <errno.h>

#include "config.h"
#include "discid.h"
#include "config_api.h"

#ifdef LINUX
#include <linux/cdrom.h>
#include "linux.h"

#elif BSD
#include <sys/cdio.h>
#include "bsd.h"

#elif SUN
#include <sys/cdio.h>
#include "sun.h"
#endif

#include "taksignal.h"
#include "cdflags.h"

#define EXIT_INVALID_TRACK 0x7F

extern char *cdrom_device;
int malloced_cdrom_device = 0;

int parse_track(char *trackname);
static void sig_term(int signo);
static void sig_tstp(int signo);
static void sig_cont(int signo);
static void sig_int(int signo);

static int print_usage(FILE *stream);
static int process_config();
int	gen_default_config();
char*	gen_conffile_name();


int	seconds=0,
	cdrom_fd;
time_t	starttime,
		stoptime;

enum {MODE_UNDEFINED, MODE_DISCID, MODE_TRACK};

int main(int argc, char *argv[]){
	int returnvalue=EXIT_SUCCESS;
	int track;

/* OLD SIGNAL HANDLING
 * if(signal(SIGTERM,sig_term)==SIG_ERR)
		fprintf(stderr,"Can't catch SIGTERM\n");
	if(signal(SIGTSTP,sig_tstp)==SIG_ERR)
		fprintf(stderr,"Can't catch SIGTSTP\n");
	if(signal(SIGCONT,sig_cont)==SIG_ERR)
		fprintf(stderr,"Can't catch SIGCONT\n");
	if(signal(SIGINT,sig_int)==SIG_ERR)
		fprintf(stderr,"Can't catch SIGINT\n");
*/

//NEW SIGNAL HANDLING (mmmm...shiny)
	struct sigaction	termaction,
						tstpaction,
						contaction,
						intaction;
	int c;
	int mode = MODE_TRACK;
	int ret;

	termaction.sa_handler=sig_term;
	termaction.sa_flags=SA_NODEFER;	//NODEFER lets signals be caught from within other handlers

	tstpaction.sa_handler=sig_tstp;
	tstpaction.sa_flags=SA_NODEFER;

	contaction.sa_handler=sig_cont;
	contaction.sa_flags=SA_NODEFER;

	intaction.sa_handler=sig_int;
	intaction.sa_flags=SA_NODEFER;

	if(-1==sigaction(SIGTERM,&termaction,NULL))
		fprintf(stderr,"Can't catch SIGTERM\n");
	if(-1==sigaction(SIGTSTP,&tstpaction,NULL))
		fprintf(stderr,"Can't catch SIGTSTP\n");
	if(-1==sigaction(SIGCONT,&contaction,NULL))
		fprintf(stderr,"Can't catch SIGCONT\n");
	if(-1==sigaction(SIGINT,&intaction,NULL))
		fprintf(stderr,"Can't catch SIGINT\n");

	ret = process_config();
	if (-1 == ret) {
		if(ENOENT==errno){
			fprintf(stderr,"Generating default config file\n");
			gen_default_config();
			ret=process_config();
			if(-1==ret){
				fprintf(stderr, "takcd: cannot process configuration file: %s\n",strerror(errno));
				return EXIT_FAILURE;
			}//if
		} else {
			fprintf(stderr, "takcd: cannot process configuration file: %s\n",strerror(errno));
			return EXIT_FAILURE;
		}//else
	}//if

	while ((c = getopt(argc, argv, "hvdc:")) != -1) {
		switch (c) {
		case 'd':
			mode = MODE_DISCID;
			break;
		case 'h':
			print_usage(stdout);
			return EXIT_SUCCESS;
		case 'v':
			printf("Tak!CD v%s - http://bard.sytes.net/takcd/\n",VERSION);
			return EXIT_SUCCESS;
		case 'c':
			if (malloced_cdrom_device) {
				free(cdrom_device);
				malloced_cdrom_device = 0;
			}//if
			cdrom_device = optarg;
			break;
		}//switch
	}//while

	if (cdrom_device == NULL){
		cdrom_device = DEV_CDROM;
	}//if

	cdrom_fd = open(cdrom_device, O_RDONLY);
	if (cdrom_fd == -1) {
		fprintf(stderr, "takcd: %s: %s\n", cdrom_device, strerror(errno));
		return EXIT_FAILURE;
	}//if

	if (mode == MODE_DISCID) {
		Discid_Init();
		discid();
		returnvalue = 0;
	} else if (mode == MODE_TRACK) {
		while (optind < argc) {
			if ((track = parse_track(argv[optind++])) < 0) {
				returnvalue = EXIT_INVALID_TRACK;
				fprintf(stderr, "Invalid track specification\n");
			} else if (play_track(cdrom_fd, track, &seconds, &starttime) < 0) {
				fprintf(stderr,"Unable to play track %d on device %s\n", track, cdrom_device);
				returnvalue = EXIT_FAILURE;
			}//else
		}//while
	}//else

	if (malloced_cdrom_device){
		free(cdrom_device);
	}//if

	close(cdrom_fd);
	
	return returnvalue;
}//main

static int print_usage(FILE *stream)
{
	return fputs("Usage: takcd [OPTION]... [TRACKSPEC]...\n"
		     "Play TRACKSPEC from a CDROM device (default all).\n\nOptions:\n"
		     " -v         display Tak!CD version information\n"
		     " -d         display disc ID\n"
		     " -c DEVICE  use DEVICE as the CDROM device instead of the platform default\n\n"
		     "TRACKSPEC is of the form trackN.cda where N is the track number\n", stream);
}

static void sig_term(int signo){
	stop_cd(cdrom_fd);
}//sig_term


static void sig_tstp(int signo){
	pause_cd(cdrom_fd,&stoptime);
}//sig_tstp


static void sig_cont(int signo){
	resume_cd(cdrom_fd,&seconds,&starttime,stoptime);
}//sig_cont

static void sig_int(int signo){
	stop_cd(cdrom_fd);
}//sig_int


int parse_track(char *trackname){
	char temp[BUFSIZ];
	int i,j=0,returnvalue=0;

	for(i=0;i<strlen(trackname);i++)
		if(isdigit((int)trackname[i]))
			temp[j++]=trackname[i];

	if(!strncmp(temp,"",BUFSIZ))
		returnvalue=-1;
	else returnvalue=atoi(temp);

	return returnvalue;
}//parse_track



static int process_config() {
	FILE *input=NULL;
	char *home_var=NULL, *file=NULL;
	config_t config;
	int saved_errno=0,
		config_opened=1;
	char cdrom_option[1024];
	int ret =  -1;
	
	if(NULL!=(file=gen_conffile_name())){
//		fprintf(stderr,"Using %s\n",file);
		if (! config_open(&config, file, C_READ)) {
			config_opened=0;
			saved_errno = errno; 
			fprintf(stderr,"Unable to open %s\n",file);
		} else	if (! config_read(&config, "devices", "cdrom", cdrom_option, sizeof(cdrom_option))) {
			saved_errno = errno;
			fprintf(stderr,"Unable to read %s\n",file);
//			goto err_read;
		} else if(NULL == (cdrom_device = strdup(cdrom_option))){
			saved_errno = errno;
//			goto err_strdup;
		} else {
			malloced_cdrom_device = 1;
			errno=saved_errno;
			ret = EXIT_SUCCESS;
		}//else
	}//if
	
	if(config_opened){
		config_close(&config);
	}//if
	free(file);
	return ret;
}//process_config

int	gen_default_config(){
	char *conffile=gen_conffile_name();
	int ret=EXIT_FAILURE;
	config_t	x;
	
	if(NULL!=conffile){
		if(!config_open(&x,conffile,C_WRITE)){	
			return ret;			
		}else if(!config_write(&x,"devices","cdrom",DEV_CDROM)){
			fprintf(stderr,"Error writing %s: %s\n",conffile,strerror(errno));
		} else {
			ret=EXIT_SUCCESS;
		}//else
		config_close(&x);
		free(conffile);
	}//if
	
	return ret;
}//gen_default_config

/* returned pointer MUST be freed! */
char*	gen_conffile_name(){
	char	*home_var=NULL,
				*file=NULL;
		
	home_var = getenv("HOME");
	if (home_var != NULL) {
		file = malloc(strlen(home_var) + 1 + /* /.takcd */ 7);
		if(NULL!=file){
			strcpy(file, home_var);
			strcat(file, "/.takcd");
		}//if
	}//if
		
	return file;
}//gen_conffile_name

/*
int play_track(int track){
	int returnvalue=0;
	struct cdrom_tocentry toc;
	struct cdrom_msf msf;

	printf("Playing track %d\n",track);

	toc.cdte_format=CDROM_MSF;
	toc.cdte_track=track;

	if((ioctl(cdrom_fd,CDROMREADTOCENTRY,&toc))<0){
		fprintf(stderr,"Error reading table of contents\n");
		returnvalue=-1;
	}//if

	msf.cdmsf_min0=toc.cdte_addr.msf.minute;
	msf.cdmsf_sec0=toc.cdte_addr.msf.second;
	msf.cdmsf_frame0=toc.cdte_addr.msf.frame;
	toc.cdte_track++;

	if((ioctl(cdrom_fd,CDROMREADTOCENTRY,&toc))<0){
		fprintf(stderr,"Error reading table of contents\n");
		returnvalue=-1;
	}//if
	else{
		msf.cdmsf_min1=toc.cdte_addr.msf.minute;
		msf.cdmsf_sec1=toc.cdte_addr.msf.second;
		msf.cdmsf_frame1=toc.cdte_addr.msf.frame;

		seconds=(msf.cdmsf_min1-msf.cdmsf_min0)*60+(msf.cdmsf_sec1-msf.cdmsf_sec0)+1;
		starttime=time(NULL);
		
		if(ioctl(cdrom_fd,CDROMPLAYMSF,&msf)==-1){
			fprintf(stderr,"Error playing track %d\n");
			returnvalue=-1;
		}//if

		printf("Track %d: %ld seconds\n",track,seconds);
		
		sleep(seconds);
	}//else

	return returnvalue;
}//play_track



int stop_cd(){
	int returnvalue=0;

	if(ioctl(cdrom_fd,CDROMSTOP,NULL)==-1){
		fprintf(stderr,"Error stopping CD\n");
		returnvalue=-1;
	}//if

	return returnvalue;
}//stop_cd
		

int resume_cd(){
	int returnvalue=0;

	seconds-=(time(NULL)-starttime);
	if(ioctl(cdrom_fd,CDROMRESUME,NULL)==-1){
		fprintf(stderr,"Error resuming cd play\n");
		returnvalue=-1;
	}//if
	else sleep(seconds);
	
	return returnvalue;
}//resume_cd


int pause_cd(){
	int returnvalue=0;

	if(ioctl(cdrom_fd,CDROMPAUSE,NULL)==-1)
		returnvalue=-1;

	pause();

	return returnvalue;
}//pause_cd
*/
